from SemiSegAPI.semiSupervised import *

__all__=['dataDistillation','modelDistillation','modelDataDistillation']

dataDistillation.__module__='distillation'
modelDistillation.__module__='distillation'
modelDataDistillation.__module__='distillation'